#####################################################################
#
#  Hornet SQL Bridge (C) 2005 Mooball.com pty ltd
#  A Zope product to mount an SQL Database into the ZODB
#
#  License: ZPL http://www.zope.org/Resources/ZPL
#  Contact: hornet@mooball.com
#
#  Please see http://plone.org/products/hornet
#
#  *WARNING: This product is ALPHA code. Do not expect it to work.*
#
#####################################################################

#####################################################################
#
#    RowSource
#    Author: James Davies (hornet@mooball.com)
#
#    Lazy wrapper around the standard python cursor class to support 
#    slicing and multiple iterators
#
#####################################################################

from AccessControl import ClassSecurityInfo
from Globals import InitializeClass


def buildDict(row, fields):
        """ 
            Builds a dictionary out of the specified row, based on self.fields 
            (The query description from the cursor)
            () => {}        
            
            Fixme: This is a significant performance hit; Could probably do this lazily;
        """
        d_row = {}
        

        fields = iter( fields )

        # Grab the next value and the corrosponding field name, and map them into the dictionary
        for value in row:
            key = fields.next()
            
            if type(key) is dict:
                key = key['name']
            
            d_row[key] = value

        return d_row    

class RowSource:
    """ 
        Wraps the C cursor object in a python iterable object
        This object keeps its own copy of the iterator position and explicitly calls scroll()
        on the cursor for each request; This allows multiple RowSource objects
        to work on the same cursor without conflicting;

        The extra scroll() method should not slow mysql down significantly.
        In most cases, we would already be on the specified row and as such 
        the scroll() call would be ignored.      
    """
    
    security = ClassSecurityInfo()
    security.setDefaultAccess('allow')
    
    def __init__(self, cursor, start=0, limit=None, position=0):
        
        
        if limit is None: limit = cursor.rowcount
            
        # Store the C Cursor
        self.cursor = cursor

        # Parse the field names from the cursors 'description'
        self.fields = [ field[0] for field in self.describe() ]

        # Set the start and end of this slice (Defaults to [:])
        self.start    = start       
        self.limit    = limit

        # Sets the current position in this iterator relative to the start of the slice
        self.position = position


    def __getslice__(self, start, end):
        """
            Return a new RowSource that iterates only over the specified slice
            If self has already been sliced, return a slice of that slice

            i,j => [i:j]
        """

        if end < 0: end = (self.limit - self.start) - (end+1)

        return RowSource( self.cursor, self.start + start, self.start + end ) 

    def __getitem__(self, item):
        """ 
            Returns the specified row
            int => {}
        """

        # Allow searching from the end of the collection (i.e. rs[-1] would return the last item)
        if item < 0: item = (self.limit - self.start) - (item+1)

        # Scroll to the specified position
        self.cursor.scroll(self.start + int(item), 'absolute')

        # Build a dictionary and return it
        return buildDict( self.cursor.fetchone(), self.fields )


    def __iter__(self):        
        """ 
            Returns a new instance of RowSource
            If we have previously been 'sliced' (i.e. result[100:200] ) 
            the resulting iterator will also only iterate over that slice
            (Multiple instances of RowSource can safely be used without conflicts on the same cursor)
        """
        return RowSource( self.cursor, self.start, self.limit )

    def next(self):
        """ 
            Seeks to the correct row, builds a dictionary of key=>value pairs, and returns it 
        """

        # Check if we are the end of this slice. If so, inform the caller we are finished iterating. 
        if( self.position > (self.limit - self.start)):
            raise StopIteration

        # Attempt to scroll to the specified position
        # If we get an IndexError then we have scrolled past the end
        #    - so inform the caller we are finished iterating
        try:
                self.cursor.scroll(self.start + self.position, 'absolute')
        except IndexError:
                raise StopIteration

        # Grab the result
        row = self.cursor.fetchone()

        # Update our internal position counter
        self.position += 1

        # Create a dictionary out of the row, and return it.
        return buildDict( row, self.fields )
        


    def seek(self, row):
        """ Lazily seeks to the specified row """
        self.position = row


    def describe(self):
        """ Returns resultset field information """
        return self.cursor.description
 
    def __len__(self):
        return int(self.limit) - int(self.start)


InitializeClass(RowSource)