#####################################################################
#
#  Hornet SQL Bridge (C) 2005 Mooball.com pty ltd
#  A Zope product to mount an SQL Database into the ZODB
#
#  License: ZPL http://www.zope.org/Resources/ZPL
#  Contact: hornet@mooball.com
#
#  Please see http://plone.org/products/hornet
#
#  *WARNING: This product is ALPHA code. Do not expect it to work.*
#
#####################################################################

#####################################################################
#
#    SQLObject
#    Author: James Davies (hornet@mooball.com)
#
#    Transient Object that maps to one row from a Table/Query
# 
#####################################################################



# Used to check default return values;
marker = '_MARKER'


# We need to use reflection to check for overloaded property methods. (i.e. _set_foo).
# However, reflection is very slow. 
# The solution is to cache a reference to the method the first time we resolve it
# (Or record the fact we couldn't find it).
# This is shared between all class instances, so it should only ever need to be done
# once per zope restart. (The logic is performed in get() and __setattr__)
_getMethod_hits = {}
_getMethod_miss = {}
_setMethod_hits = {}
_setMethod_miss = {}

class SQLObject:    
    
    def __init__(self, data=None, id=None):
        if not self.__class__ in _getMethod_hits:
            self.resetOverloadCache()
            
        self._data = data  # Store initial version. 
        self._delta = {}   # Dictionary to store deltas of changed information
        
    ###########################################################################    
                    
    def resetOverloadCache(self):
        """ Resets the cache of overloaded set/get methods """
        _getMethod_hits[self.__class__] = {} 
        _getMethod_miss[self.__class__] = {}
        _setMethod_hits[self.__class__] = {}
        _setMethod_miss[self.__class__] = {} 
    
    ###########################################################################    

    def __getattr__(self, key, default=marker):
        """ Attempts to retrieve the specified key from this object """
        if not key.startswith('_'):
            result = self.get(key, marker)
            if result is not marker:
                    return result 
        
        # Then look for a standard python attribute
        value = self.__dict__.get(key, marker)
        if value is not marker: return value
        
        if default is not marker:
            return default
        
        raise AttributeError(key)
    
    ###########################################################################    
    
    def get(self, key, default=None):
        """ Get the specified key from this objects SQL record """
        
        hits = _getMethod_hits[self.__class__]
        miss = _getMethod_miss[self.__class__]
        
        # Look for a custom get method
        methodName = '_get_%s' % key
        if methodName in hits:
            return hits[methodName](self)
        
        if methodName not in miss and hasattr(self.__class__, methodName): 
                hits[methodName] = getattr(self.__class__, methodName)
                return hits[methodName](self)
        
        if methodName not in miss:
            miss[methodName] = True
        
        # Look in our SQL Record
        if self._data:
            #value = self._data.get(key, marker)
            value = self.getValue(key, marker)            
            if value is not marker: return value
        
        return default
        
    def getValue(self, key, default=None):
        if key in self._data:
                v = self._data[key]
                if type(v) is str and hasattr(self._v_mountpoint, 'recode'):
                    recode = getattr(self._v_mountpoint, 'recode')
                    eFrom, eTo = recode.split('|')
                    v = v.decode(eFrom).encode(eTo)                    
                    
                return v
                    
                        
        else:
                return default

    def __setattr__(self, key, value):                 
        # Sets the specified key/value pair on this object
    
        # Look for an overloadded method; if it exists, call it.
        # The method should return True if it changed this object
        
        if key.startswith('_'):
            self.__dict__[key] = value
        else:            
        
            hits = _setMethod_hits[self.__class__]
            miss = _setMethod_miss[self.__class__]
            
            methodName = "_set_%s" % key
            
            if methodName in hits:
                if hits[methodName](self,value):
                    self._register()
                    self._p_changed = 1
                    return
            
            if methodName not in miss and methodName in dir(self):
                    hits[methodName] = getattr(self.__class__, methodName) 
                    if hits[methodName](self,value):
                        self._register()
                        self._p_changed = 1
                        return
                        
            # Mutator method was not found; manually add to dictionary
            miss[methodName] = True
        
            data  = self._data
            delta = self._delta 
        
            if data.has_key(key):
                self.setValue(key, value)
            else:
                self.__dict__[key] = value
        
            
                        
            
    def setValue(self, key, value):
        """
            Directly set the specified key:value pair into our data dictionary, ignoring 
            the advanced __setattr__ logic 
        """
        
        self._data[key]  = value
        self._delta[key] = value
        self._register()
        self._p_changed = 1
            
    def set(self, key, value):
        """
            If the key exists in our schema, record in our data dictionary.
            else, set as an attribute of the transient object (will not be saved)
        """            
        setattr(self, key, value)
        
    def sql_commit(self):
        """ Commits this object. Is called automatically when zope tells us to commit """
        self.save()
        
    def save(self):
        primaryKey = self._v_folder.getPrimaryKey()
        
        # If the primary key is set (the record exists) then create an sql update statement
        #if getattr( self, primaryKey, None ):
        if self.isExisting():
            self.updateDatabase()
            self._delta.clear() 
        # If the primary key isn't set (the record doesn't exist), create a new one
        else:            
            insert_id = self.insertIntoDatabase()
            self._data[primaryKey] = insert_id
            self._delta.clear()
            return insert_id
        
    def updateDatabase(self):
        primaryKey   = self._v_folder.getPrimaryKey()
        table        = self._v_folder.getTableName()
            
        primaryValue  = self.getPrimaryValue()
        
        #if primaryKey in self._delta: del self._delta[primaryKey]
        
        self._v_mountpoint.sqlUpdate( table, self._delta, primaryKey, primaryValue )
    
    def insertIntoDatabase(self):
        primaryKey   = self._v_folder.getPrimaryKey()
        table        = self._v_folder.getTableName()
        
        #if primaryKey in self._data: del self._data[primaryKey]
        
        insert_id = self._v_mountpoint.sqlInsert( table, self._data )
        
        self._new = False
        
        return insert_id
                
        
    def isExisting(self):
        """ Does the object already exist? """
        
        # Is the object explicitly marked as new? (i.e. created by createEmty()?)
        if getattr(self, '_new', False):
            return False # ... if so, return False
        
        primaryKey   = self._v_folder.getPrimaryKey()
        
        id = self.getPrimaryValue()
        if id == 'create': return False
        return id and True or False
        
        # If the primary key is set, then we exist.
        #return self.data.get('primaryKey') and True or False
            
        #return getattr( self, primaryKey, None ) and True or False
        
        
        
    def getPrimaryValue(self):
        return self._data[self._v_folder.getPrimaryKey()]
        
        
    def keys(self):
        return self._data.keys()
        
    def Primary(self):
        """ Returns the primary key for this object """
        return self.primary
    
    def Table(self):
        """ Returns the table for this object """
        return self.table        
        
    def getSchema(self):
        """ Returns the schema for this object """
        if hasattr(self, 'cb_schema'):
            cb_schema_name = self.cb_schema
            cb_schema      = getattr(self, cb_schema_name)
            return cb_schema()
        
            
